'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var redux = require('redux');
var memoizeOne = _interopDefault(require('memoize-one'));
var produce = _interopDefault(require('immer'));
var thunk = _interopDefault(require('redux-thunk'));

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

var effectSymbol = Symbol('effect');
var selectSymbol = Symbol('select');
var selectDependeciesSymbol = Symbol('selectDependencies');
var selectStateSymbol = Symbol('selectState');

var isObject = function isObject(x) {
  return x && _typeof(x) === 'object' && !Array.isArray(x);
};

var get = function get(path, target) {
  return path.reduce(function (acc, cur) {
    return isObject(acc) ? acc[cur] : undefined;
  }, target);
};

var set$1 = function set(path, target, value) {
  path.reduce(function (acc, cur, idx) {
    if (idx + 1 === path.length) {
      acc[cur] = value;
    } else {
      acc[cur] = acc[cur] || {};
    }

    return acc[cur];
  }, target);
};

var effect = function effect(fn) {
  // eslint-disable-next-line no-param-reassign
  fn[effectSymbol] = true;
  return fn;
};
var select = function select(fn, dependencies) {
  var selector = memoizeOne(function (state) {
    return fn(state);
  });
  selector[selectSymbol] = true;
  selector[selectDependeciesSymbol] = dependencies;
  selector[selectStateSymbol] = {};
  return selector;
};
var createStore = function createStore(model) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var _options$devTools = options.devTools,
      devTools = _options$devTools === void 0 ? true : _options$devTools,
      _options$middleware = options.middleware,
      middleware = _options$middleware === void 0 ? [] : _options$middleware,
      _options$initialState = options.initialState,
      initialState = _options$initialState === void 0 ? {} : _options$initialState;

  var definition = _objectSpread({}, model, {
    logFullState: function logFullState(state) {
      // eslint-disable-next-line no-console
      console.log(JSON.stringify(state, null, 2));
    }
  });

  var references = {};
  var defaultState = {};
  var actionEffects = {};
  var actionCreators = {};
  var actionReducers = {};
  var selectorReducers = [];

  var extract = function extract(current, parentPath) {
    return Object.keys(current).forEach(function (key) {
      var value = current[key];

      var path = _toConsumableArray(parentPath).concat([key]);

      if (typeof value === 'function') {
        var actionName = path.join('.');

        if (value[selectSymbol]) {
          // skip
          value[selectStateSymbol] = {
            parentPath: parentPath,
            key: key,
            executed: false
          };
          selectorReducers.push(value);
        } else if (value[effectSymbol]) {
          // Effect Action
          var action = function action(payload) {
            references.dispatch({
              type: actionName,
              payload: payload
            });
            return value(references.dispatch, payload, {
              getState: references.getState
            });
          };

          action.actionName = actionName;
          set$1(path, actionEffects, action); // Effect Action Creator

          set$1(path, actionCreators, function (payload) {
            return references.dispatch(function () {
              return Promise.resolve(action(payload));
            });
          });
        } else {
          // Reducer Action
          var _action = function _action(state, payload) {
            return produce(state, function (draft) {
              return value(draft, payload, {
                dispatch: references.dispatch,
                dispatchLocal: get(path, references.dispatch),
                getState: references.getState
              });
            });
          };

          _action.actionName = actionName;
          set$1(path, actionReducers, _action); // Reducer Action Creator

          set$1(path, actionCreators, function (payload) {
            return references.dispatch({
              type: _action.actionName,
              payload: payload
            });
          });
        }
      } else if (isObject(value) && Object.keys(value).length > 0) {
        extract(value, path);
      } else {
        // State
        var initialParentRef = get(parentPath, initialState);

        if (initialParentRef && key in initialParentRef) {
          set$1(path, defaultState, initialParentRef[key]);
        } else {
          set$1(path, defaultState, value);
        }
      }
    });
  };

  extract(definition, []);

  var createReducers = function createReducers(current, path) {
    var actionReducersAtPath = Object.keys(current).reduce(function (acc, key) {
      var value = current[key];

      if (typeof value === 'function' && !value[effectSymbol]) {
        return _toConsumableArray(acc).concat([value]);
      }

      return acc;
    }, []);
    var stateAtPath = Object.keys(current).reduce(function (acc, key) {
      return isObject(current[key]) ? _toConsumableArray(acc).concat([key]) : acc;
    }, []);
    var nestedReducers = stateAtPath.map(function (key) {
      return [key, createReducers(current[key], _toConsumableArray(path).concat([key]))];
    });

    var reducerForActions = function reducerForActions() {
      var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : get(path, defaultState);
      var action = arguments.length > 1 ? arguments[1] : undefined;
      var actionReducer = actionReducersAtPath.find(function (x) {
        return x.actionName === action.type;
      });

      if (actionReducer) {
        return actionReducer(state, action.payload);
      }

      for (var i = 0; i < nestedReducers.length; i += 1) {
        var _nestedReducers$i = _slicedToArray(nestedReducers[i], 2),
            key = _nestedReducers$i[0],
            reducer = _nestedReducers$i[1];

        var newState = reducer(state[key], action);

        if (state[key] !== newState) {
          return _objectSpread({}, state, _defineProperty({}, key, newState));
        }
      }

      return state;
    };

    var isInitial = true;

    var runSelectorReducer = function runSelectorReducer(state, selector) {
      var _selector$selectState = selector[selectStateSymbol],
          parentPath = _selector$selectState.parentPath,
          key = _selector$selectState.key,
          executed = _selector$selectState.executed;

      if (executed) {
        return state;
      }

      var dependencies = selector[selectDependeciesSymbol];
      var newState = produce(dependencies ? dependencies.reduce(runSelectorReducer, state) : state, function (draft) {
        // eslint-disable-next-line no-param-reassign
        var target = parentPath.length > 0 ? get(parentPath, draft) : draft;

        if (target) {
          target[key] = selector(target);
        }
      }); // eslint-disable-next-line no-param-reassign

      selector[selectStateSymbol].executed = true;
      return newState;
    };

    var runSelectors = function runSelectors(state) {
      return selectorReducers.reduce(runSelectorReducer, state);
    };

    return selectorReducers.length ? function (state, action) {
      var stateAfterActions = reducerForActions(state, action);

      if (state !== stateAfterActions || isInitial) {
        var stateAfterSelectors = runSelectors(stateAfterActions);
        isInitial = false;
        selectorReducers.forEach(function (selector) {
          // eslint-disable-next-line no-param-reassign
          selector[selectStateSymbol].executed = false;
        });
        return stateAfterSelectors;
      }

      return stateAfterActions;
    } : reducerForActions;
  };

  var reducers = createReducers(actionReducers, []);
  var composeEnhancers = devTools && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ : redux.compose;
  var store = redux.createStore(reducers, defaultState, composeEnhancers(redux.applyMiddleware.apply(void 0, [thunk].concat(_toConsumableArray(middleware))))); // attach the action creators to dispatch

  Object.keys(actionCreators).forEach(function (key) {
    store.dispatch[key] = actionCreators[key];
  });
  references.dispatch = store.dispatch;
  references.getState = store.getState;
  return store;
};

exports.createStore = createStore;
exports.effect = effect;
exports.select = select;
//# sourceMappingURL=easy-peasy.js.map
